@@ -116,9 +116,20 @@ static void cache_invalidate_dir(const char *path)
     pthread_mutex_unlock(&cache.lock);
 }
 
+static int cache_del_children(const char *key, void *val_, const char *path)
+{
+    (void) val_;
+    if (strncmp(key, path, strlen(path)) == 0)
+        return TRUE;
+    else
+        return FALSE;
+}
+
 static void cache_do_rename(const char *from, const char *to)
 {
     pthread_mutex_lock(&cache.lock);
+    g_hash_table_foreach_remove(cache.table, (GHRFunc) cache_del_children,
+                                (char *) from);
     cache_purge(from);
     cache_purge(to);
     cache_purge_parent(from);
@@ -250,12 +261,14 @@ static int cache_dirfill(fuse_cache_dirh_t ch, const char *name,
                          const struct stat *stbuf)
 {
     int err = ch->filler(ch->h, name, 0, 0);
-    if (!err && (stbuf->st_mode & S_IFMT)) {
-        char *fullpath;
+    if (!err) {
         g_ptr_array_add(ch->dir, g_strdup(name));
-        fullpath = g_strdup_printf("%s/%s", !ch->path[1] ? "" : ch->path, name);
-        cache_add_attr(fullpath, stbuf);
-        g_free(fullpath);
+        if (stbuf->st_mode & S_IFMT) {
+            char *fullpath =
+                g_strdup_printf("%s/%s", !ch->path[1] ? "" : ch->path, name);
+            cache_add_attr(fullpath, stbuf);
+            g_free(fullpath);
+        }
     }
     return err;
 }
